'use strict';

const semver = require('semver');
const fs = require('fs');
const os = require('os');
const pretag = 'pre_';
const consolidateMethod = 'Consolidate';

module.exports = {

  addons: {

    _semverAskOptions(options, config) {
      const list = {
        type: 'list',
        name: '_strategy',
        required: true,
        choices: [],
        message: 'Choose next release change'
      };
      Object.getOwnPropertyNames(options).forEach((name) => {
        list.choices.push({name: name, value: options[name]});
      });
      this.params.promptOptions = [ list ];
      return this.inquire('promptOptions')
        .then(() => {
          config.strategyName = this.params._strategy[0];
          config.prereleaseTag = this.params._strategy.length > 1 ? this.params._strategy[1] : undefined;
          return Promise.resolve(config);
        });
    },

    _semverLastTag(config) {
      return this.execute(this.gitCmd, ['tag', '-l'])
        .then((result) => {
          if (result.output) {
            const lastVersion = result.output.split(os.EOL)
              .map((tag) => {
                if (config.prereleaseTag && config.pretag) {
                  return tag.indexOf(config.prereleaseTag) >= 0 ? tag.replace(pretag, '') : undefined;
                } else if (!tag.startsWith(pretag) && config.strategyName === 'auto' ? !tag.indexOf('-') : true) {
                  return tag;
                }
              })
              .filter((tag) => tag && semver.valid(tag))
              .sort((a, b) => {
                return semver.gt(a, b) ? 1 : -1;
              })
              .pop();
            if (lastVersion) {
              this.logger.info('last pretag version:', lastVersion);
              return lastVersion;
            }
          }
        });
    },

    _semverOptions(options, config) {
      if (options) {
        if (Object.prototype.toString.call(options) !== '[object Array]') {
          return this._semverAskOptions(options, config);
        } else {
          config.strategyName = options[0];
          config.prereleaseTag = options.length > 1 ? options[1] : undefined;
          return Promise.resolve(config);
        }
      } else {
        return Promise.resolve(config);
      }
    },

    _semverLastVersion(config) {
      return this._semverLastTag(config)
        .then((lastTagVersion) => {
          if (lastTagVersion && semver.gte(lastTagVersion.split('-')[0], config.value.version.split('-')[0])) {
            config.value.version = lastTagVersion;
          }
          return config;
        });
    },

    semverStrategy(config) {
      if (config && config.strategyName) {
        return this._semverLastVersion(config)
          .then((_config) => {
            this.logger.info('version:', _config.value.version, 'strategy:', _config.strategyName, 'tag:', _config.prereleaseTag);
            switch (_config.strategyName) {
              case 'timestamp':
                _config.value.version = `${_config.value.version}.${new Date().toISOString()}`;
                break;
              case 'remove':
                _config.value.version = _config.value.version.replace(`${_config.prereleaseTag}.`, '');
                _config.consolidate = true;
                break;
              case 'auto':
                if (semver.gt(_config.initialVersion.split('-')[0], _config.value.version.split('-')[0])) {
                  _config.value.version = _config.initialVersion.split('-')[0];
                } else {
                  let strategy = semver.diff(_config.value.version.split('-')[0], _config.initialVersion.split('-')[0]);
                  strategy = strategy ? strategy : 'patch';
                  this.logger.info('Auto strategy using', '#green', strategy);
                  _config.value.version = semver.inc(_config.value.version, strategy);
                }
                _config.consolidate = true;
                break;
              default:
                _config.value.version = semver.inc(_config.value.version, _config.strategyName, _config.prereleaseTag);
            }
            return _config;
          });
      } else {
        this.logger.warn('No strategy configured!!');
        return Promise.resolve();
      }
    },

    semverConsolidate(config) {
      const tool = config.name.split('.')[0];
      const func = this[`${tool}${consolidateMethod}`];
      if (func) {
        this.logger.info('Executing consolidate shrinkwrap on ', '#green', tool);
        return func(config)
          .then((_config) => {
            if (_config.hasChanged) {
              return this.scmAdd({file: _config.hasChanged})
                .then(() => this.scmCommit({message: `chore(): update ${_config.hasChanged}`, file: _config.hasChanged}))
                .then(() => _config);
            } else {
              return _config;
            }
          });
      } else {
        this.logger.info('No method', '#cyan', `${tool}${consolidateMethod}`, 'is on any plugin configurated!');
        return Promise.reject({error: `No ${tool}${consolidateMethod} method is present on any plugin configured`});
      }
    },

    semverSaveVersion(config) {
      if (config) {
        const changed = config.value.version && config.initialVersion !== config.value.version;
        this.logger.info('Final version on', '#cyan', config.name, '->', '#green', config.value.version, changed ? '' : '(Not changed)');
        if (changed) {
          fs.writeFileSync(config.name, JSON.stringify(config.value, null, 2));
          return this.scmCommit({message: `${this.params.versionDescription ? this.params.versionDescription : 'chore(): change version'} ${config.value.version}`, file: config.name})
            .then(() => config.consolidate ? this.semverConsolidate(config) : config)
            .then(() => config.pretag ? this.execute(this.gitCmd, ['tag', `${pretag}${config.value.version}`]) : config);
        }
      } else {
        this.logger.warn('No config object for', '#red', 'IS NOT POSSIBLE TO SAVE VERSION');
      }
      return Promise.resolve();
    },

    semverGetVersionConfig(preVersion) {
      let count = 0;
      const config = this.params.jsonFiles
        .map((file) => {
          if (this.fsExists(file)) {
            const configJson = this.fsReadConfig(file);
            if (configJson.version) {
              if (semver.valid(configJson.version)) {
                this.logger.info('Initial version on', '#cyan', file, '->', '#green', configJson.version);
                count++;
                const result = {
                  name: file,
                  value: configJson,
                  initialVersion: configJson.version
                };
                result.value.version = preVersion ? preVersion : result.value.version;
                return result;
              } else {
                this.logger.error('Initial version on', '#cyan', file, '->', configJson.version, '#red', 'not valid!');
              }
            } else {
              this.logger.error('File', '#green', file, 'has no version!!');
            }
          }
        })
        .filter((e) => e)
        .pop();
      if (count > 1) {
        this.logger.warn('#yellow', 'WARNING:', 'There is more than one version file, taking last file!!!');
      }
      return config;
    },

    semverConfig(options, usePretag, preVersion) {
      const config = this.semverGetVersionConfig(preVersion);
      if (config) {
        config.pretag = usePretag;
        return this._semverOptions(options, config);
      } else {
        this.logger.warn('There is no version json file, available list:', '#cyan', this.params.jsonFiles.join(', '));
        return Promise.resolve();
      }
    }
  }

};